# SPDX-License-Identifier: GNU GPL v3
"""
A class for building an MDP environment of image filters for graph generation.
"""



class SGTGraphEnv:
    """
    Class for building a (Markov Decision Process) MDP environment for manipulating the Genetic Algorithm (GA) optimizer
    into navigating the search space of image filters.
    """

    def __init__(self):
        pass

    def _build_action_space(self):
        pass

    def _build_observation_space(self):
        pass
